import matplotlib.pyplot as plt
import numpy as np

file  = "cifar10_dataset.npz"
url  = "https://github.com/danielwilczak101/EasyNN/raw/datasets/cifar/cifar10_dataset.npz"
labels = {
            0: "airplane",
            1: "automobile",
            2: "bird",
            3: "cat",
            4: "deer",
            5: "dog",
            6: "frog",
            7: "horse",
            8: "ship",
            9: "truck"
}

def show(image):
    """Show the image as an RGB image.
        Args:
            image: Numpy array of an image structured in RGB and flattened.
        
        Return:
            image: [32,32] image of the RGB stacked on top of itself.
        Example:
            >>> show(image[0])
            Will show a matplot lib of the a [32,32] image.
    """
    
    im_r = image[0:1024].reshape(32, 32)
    im_g = image[1024:2048].reshape(32, 32)
    im_b = image[2048:].reshape(32, 32)

    img = np.dstack((im_r, im_g, im_b))

    plt.imshow(img) 
    plt.show()


from dataclasses import dataclass, field
from collections import defaultdict


@dataclass
class Print:
    settings: dict[str, list[list]] = field(default_factory=lambda: defaultdict(list))

    def get_attribute(self, model, attribute: str) -> str:
        if attribute == "iteration":
            return f"Iteration: {model.training.iteration}"
        elif attribute == "validation_accuracy":
            return f"Validation Accuracy: {model.accuracy(*model.validation.sample)}"
        else:
            raise NotImplementedError
    
    def callback(self, model, setting, translation, frequency: int=1, **kwargs) -> None:
        if model is None:
            setting.append([0, frequency, [translation.get(attribute, attribute) for attribute in kwargs]])
            return
        for line in setting:
            if line[0] == 0:
                print(*[self.get_attribute(model, attribute) for attribute in line[2]], sep=", ")
            line[0] += 1
            line[0] %= line[1]

    # -------------- Training -------------- #

    def on_training_start(self, model=None, **kwargs) -> None:
        # For `model.print.on_optimization_start(validation_accuracy=True)`.
        self.callback(
            model,
            self.settings["on_training_start"],
            {"accuracy": "training_accuracy", "loss": "training_loss"},
            **kwargs
        )


    def on_training_end(self, model=None, **kwargs) -> None:
        # For `model.print.on_optimization_start(validation_accuracy=True)`.
        self.callback(
            model,
            self.settings["on_training_end"],
            {"accuracy": "training_accuracy", "loss": "training_loss"},
            **kwargs
        )
    # -------------- Validation -------------- #
    
    def on_validation_start(self, model=None, **kwargs) -> None:
        self.callback(
            model,
            self.settings["on_validation_start"],
            {"accuracy": "validation_accuracy", "loss": "validation_loss"},
            **kwargs
        )

    def on_validation_end(self, model=None, **kwargs) -> None:
        self.callback(
            model,
            self.settings["on_validation_end"],
            {"accuracy": "validation_accuracy", "loss": "validation_loss"},
            **kwargs
        )
    # -------------- Optimization -------------- #

    def on_optimization_start(self, model=None, **kwargs) -> None:
        # Doesnt print anything right now
        self.callback(
            model,
            self.settings["on_optimization_start"],
            {},
            **kwargs
        )

    def on_optimization_end(self, model=None, **kwargs) -> None:
        # Doesnt print anything right now
        self.callback(
            model,
            self.settings["on_optimization_end"],
            {},
            **kwargs
        )


from typing import Protocol


class Callback(Protocol):

    def on_optimization_start(self, model):
        pass

    def on_optimization_end(self, model):
        pass

    def on_training_start(self, model):
        pass

    def on_training_end(self, model):
        pass

    def on_validation_start(self, model):
        pass

    def on_validation_end(self, model):
        pass

    def on_testing_start(self, model):
        pass

    def on_testing_end(self, model):
        pass

    def on_epoch_start(self, model):
        pass

    def on_epoch_end(self, model):
        pass


from __future__ import annotations
import numpy as np
from typing import Generic, TypeVar
from EasyNN.model.abc import Model
from EasyNN.typing import ArrayND

ArrayShape = TypeVar("ArrayShape", bound=ArrayND)


class Randomize(Model[ArrayShape, ArrayShape], Generic[ArrayShape]):
    """Randomizes the training input for data augmentation purposes."""
    _parameters: ArrayND = np.empty(0, dtype=float)
    _derivatives: ArrayND = np.empty(0, dtype=float)
    noise: float

    def __init__(self: Randomize[ArrayShape], noise: float = 0.1) -> None:
        self.noise = noise

    def __forward__(self: Randomize[ArrayShape], x: ArrayShape) -> ArrayShape:
        if self.command.startswith("on_training_"):
            scale = self.noise / (1 + 1e-2 * self.training.iteration) ** 0.2
            return x + np.random.normal(scale=scale, size=x.shape)
        return x

    def __backward__(self: Randomize[ArrayShape], dy: ArrayShape, y: ArrayShape = None, use_y: bool = False) -> ArrayShape:
        return dy

    def __setup__(self: Normalize[ArrayShape]) -> None:
        pass



from __future__ import annotations
from typing import Generic, TypeVar, TypedDict, overload
from EasyNN.model.abc import Model
from EasyNN.typing import ArrayND
import numpy as np

ArrayShape = TypeVar("ArrayShape", bound=ArrayND)
ArrayDict = TypedDict("ArrayDict", parameters=ArrayND, mean=ArrayShape, variance=ArrayShape)


class Normalize(Model[ArrayShape, ArrayShape], Generic[ArrayShape]):
    """Normalizes the input based on the training data."""
    __lr: float
    _mean: ArrayShape = 0.0
    _variance: ArrayShape = 1e-4
    _weight: float = 0.0
    _parameters: ArrayND = np.empty(0, dtype=float)
    _derivatives: ArrayND = np.empty(0, dtype=float)

    def __init__(self: Normalize[ArrayShape], lr: float = 1e-2) -> None:
        self.__lr = lr

    @property
    def mean(self: Normalize[ArrayShape]) -> ArrayShape:
        """Returns an approximation for the mean of the provided training data."""
        return self._mean / self._weight

    @property
    def variance(self: Normalize[ArrayShape]) -> ArrayShape:
        """Returns an approximation for the variance of the provided training data."""
        return self._variance / self._weight

    @property
    def deviation(self: Normalize[ArrayShape]) -> ArrayShape:
        """Returns an approximation for the deviation of the provided training data."""
        return np.sqrt(self.variance)

    def get_arrays(self) -> ArrayDict:
        """Returns the arrays stored in the model."""
        arrays = dict()
        if hasattr(self, "parameters"):
            arrays["parameters"] = self.parameters
        if hasattr(self, "mean"):
            arrays["mean"] = self.mean
        if hasattr(self, "variance"):
            arrays["variance"] = self.variance
        return arrays

    @overload
    def set_arrays(self, *, parameters: ArrayND = None) -> None:
        ...

    @overload
    def set_arrays(self, *, parameters: ArrayND = None, mean: ArrayND, variance: ArrayND) -> None:
        ...

    def set_arrays(self, *, parameters=None, mean=None, variance=None) -> None:
        """Sets the arrays stored in the model."""
        # Set the parameters, if any.
        if parameters is not None:
            self.parameters = parameters
        if mean is None and variance is None:
            return
        elif mean is None or variance is None:
            raise TypeError("either both the mean and variance must be given or neither, not just one")
        else:
            self._mean = mean
            self._variance = variance + type(self)._variance
            self._weight = 1.0

    def __forward__(self: Normalize[ArrayShape], x: ArrayShape) -> ArrayShape:
        if self.command.startswith("on_training_"):
            self._weight += self.__lr * (1 - self._weight)
            self._mean += self.__lr * (x.mean(axis=0) - self._mean)
            self._variance += self.__lr * (((x - self.mean) ** 2).mean(axis=0) - self._variance)
        elif self._weight == 0.0:
            return x
        return (x - self.mean) / self.deviation

    def __backward__(self: Normalize[ArrayShape], dy: ArrayShape, y: ArrayShape = None, use_y: bool = False) -> ArrayShape:
        if self._weight == 0.0:
            return dy
        return dy / self.deviation

    def __setup__(self: Normalize[ArrayShape]) -> None:
        pass



from __future__ import annotations
from itertools import chain
import numpy as np
from typing import Iterator
from EasyNN.batch.abc import Batch
from EasyNN.dataset import Dataset
from EasyNN.typing import Sample


class MiniBatch(Batch):
    """Produces mini-batches, which are multiple data points at a time."""
    size: int

    def __init__(self: MiniBatch, size: int = 32) -> None:
        self.size = size

    def generate_samples(self: MiniBatch, dataset: Dataset) -> Iterator[Sample]:
        # Loop over random indexes grouped together as the batch size.
        dataset._batch_size = self.size
        random_indexes = chain.from_iterable(np.random.permutation(len(dataset)) for _ in iter(int, None))
        grouped_indexes = zip(*[random_indexes] * self.size)
        return (dataset[i,] for i in grouped_indexes)



from EasyNN.examples.cifar10.structure import model
#from EasyNN.examples.cifar10.trained import model
from EasyNN.examples.cifar10.data import dataset

# Set your models data
model.training.data = dataset



import numpy as np
from PIL import ImageOps, Image as pil_image

__all__ = ["Preprocess"]

class Preprocess:
    """Used for preprocessing images to a specific look.

    Args:
        image_path: Path to where the image is located relative to the executing file.
    """

    def __init__(self, image_path: str):
        try: 
            # We opened the image
            self.image = pil_image.open(image_path)
        except FileNotFoundError:
            raise FileNotFoundError(f"{image_path} not found.") from None

    def format(
        self,
        grayscale=False,
        invert:bool=False,
        rotate:int=0,
        contrast:int=0,
        resize:list[int]=False,
        flatten:str='C',
        process: bool=False,
        convert: str=None,
        intensity: int=0
        ):
        """Used to specify how they want the format of the image to look like.
        
        Args:
            grayscale: Used to make an RGB image into grayscale format.
            invert: Light areas are mapped to dark, and dark areas are mapped to light.
            rotate: If the image needs to be rotated by a multiple of 90 degrees.
                Ex. rotate = 1 -> rotate image 90 degrees \n
                Ex. rotate = 3 -> rotate image 270 degrees
            contrast: How much addition pixel strength you want to remove from the array.
                Ex. original = [10,20,30] if offest = 5 then new_array = [5,15,25]
            resize: Specify an a desired pixel/image size.
                Ex. resize([28,28]) -> Image will now be a 28x28 pixel image.   
        """
        if convert:
            print(convert)
            self.image.convert(convert)
        if grayscale:
            # Convert to grey scale.
            self.image = self.image.convert('L')
        if invert:
            # Inverted the color to look like our dataset images.
            self.image = ImageOps.invert(self.image)
        if resize:
            # Resize the iamge.
            self.image = self.image.resize(resize)

        # Convert to numpy array and not an image.
        self.image = np.array(self.image)
        
        if rotate:
            # Transofrm the image into a numpy array and apply rotation if needed.
            self.image = np.rot90(self.image, k=rotate)
        if process:
            # Apply contrast filter out all the grey noise in our image using the mean value.
            self.image = np.where(self.image < np.mean(self.image) + contrast,0,self.image)
        # Flatten the image so it can be used in the image and show functions.
        if flatten:
            self.image = self.image.flatten(flatten)
        if intensity:
            # Expose / brighten the pixels that are non zero.
            self.image[self.image!=0] += intensity
        
        return self.image

        


from __future__ import annotations
from typing import Any, Generic, IO, Iterator, TypeVar, Union
from EasyNN.utilities.download import download
from EasyNN.utilities.data.load import load
from EasyNN.typing import ArrayND
import EasyNN.batch.abc

ArrayIn = TypeVar("ArrayIn", bound=ArrayND)
ArrayOut = TypeVar("ArrayOut", bound=ArrayND)


class Dataset(Generic[ArrayIn, ArrayOut]):
    """
    Class for storing a dataset and looping over multiple arrays together.

    Example:
        >>> x = np.array([...])
        >>> y = np.array([...])
        >>> training = Dataset()
        >>> training.data = (x, y)
        >>> assert training[0] == (x[0], y[0])
    """
    _batch: EasyNN.batch.abc.Batch
    _batch_size: int
    _data: tuple[ArrayIn, ArrayOut]  # Don't save
    percent: float = 0.05
    file: Union[str, IO]
    url: str
    sample: tuple[ArrayIn, ArrayOut]
    samples: Iterator[tuple[ArrayIn, ArrayOut]]  # Don't save
    iteration: int = -1

    def __getstate__(self) -> dict[str, Any]:
        return {
            name.removeprefix("_") if name != "_batch_size" else "_batch_size": attribute
            for name, attribute in vars(self).items()
            if name not in ("_data", "sample", "samples")
        }

    def __getitem__(self, indexes: Union[int, tuple[int, ...]]) -> tuple[ArrayIn, ArrayOut]:
        """Get a sample from the data."""
        return tuple([x[indexes] for x in self.data])

    def __iter__(self):
        """Loops over the data in accordance to the indexes."""
        return self

    def __len__(self) -> int:
        """Returns the size of the dataset."""
        return len(self.data[0])

    def __next__(self) -> tuple[ArrayIn, ArrayOut]:
        """Update and return the next sample."""
        # Each sample corresponds to 1 iteration.
        self.iteration += 1
        self.sample = next(self.samples)
        return self.sample

    @property
    def epoch(self) -> int:
        """The epoch is how many times the whole dataset has been looped through."""
        return self.iteration * self._batch_size // len(self)

    @property
    def batch(self) -> EasyNN.batch.abc.Batch:
        """When setting the batch, get the samples using the batch."""
        return self._batch

    @batch.setter
    def batch(self, batch: EasyNN.batch.abc.Batch) -> None:
        self._batch = batch
        self._batch_size = batch.size
        self.samples = batch.generate_samples(self)

    @property
    def data(self) -> tuple[ArrayIn, ArrayOut]:
        """
        The dataset's data. If the data is not found, this first checks
        for a file storing the data, then a url to load the data from.
        """
        return self._data

    @data.setter
    def data(self, data: tuple[ArrayIn, ArrayOut]) -> None:
        # TODO: Maybe some input checking?
        self._data = data



"""
TODO: Not complete.
"""
from __future__ import annotations
from abc import ABC, abstractmethod
from itertools import count, cycle
from typing import Hashable, get_args
from EasyNN._abc import AutoDocumentation
from EasyNN.typing import Array1D, Command
import EasyNN.model.abc


class Optimizer(AutoDocumentation, ABC):
    """
    Abstract Base Class for the Optimizer.
    """
    models: dict[Hashable, EasyNN.model.abc.Model]
    lr: float

    def __init__(self: Optimizer, models: dict[Hashable, EasyNN.model.abc.Model] = None, lr: float = 1e-2) -> None:
        self.models = dict() if models is None else models
        self.lr = lr

    def add(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Add a model to the optimizer."""
        for i in count(len(self.models)):
            if i not in self.models:
                self.models[i] = model
                break

    def train(self: Optimizer, *models: EasyNN.model.abc.Model) -> tuple[EasyNN.model.abc.Model, ...]:
        """
        Train the models.

        Example:
            >>> optimizer.train(model_1, model_2)
            (model_1, model_2)
        """
        # By default, use the saved optimizer models.
        if len(models) == 0:
            models = tuple(self.models.values())
        # Setup the model commands.
        for model in models:
            self.setup(model)
        # Cycle through each model.
        for model in cycle(models):
            # Run the next command for the model and stop if there are no more.
            if next(model.commands, None) is None:
                return models

    def setup(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Setup a model."""
        # Assign learning rates.
        model._optimizer_lr = self.lr
        # Add on the optimizer commands.
        model.callback(self)
        # Begin running model commands.
        model.commands = model.optimizer_commands()

    def get_derivatives(self: Optimizer, model: EasyNN.model.abc.Model) -> Array1D[float]:
        """Computes the derivatives for the optimizer."""
        # Use the testing/validation samples.
        if model.command.startswith("on_testing"):
            model.sample_derivatives(*model.testing.sample)
        elif model.command.startswith("on_validation"):
            model.sample_derivatives(*model.validation.sample)
        # Use the training sample by default.
        else:
            model.sample_derivatives(*model.training.sample)
        return model.derivatives

    def on_optimization_start(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Ran at the start of optimization."""

    def on_optimization_end(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Ran at the end of optimization."""

    @abstractmethod
    def on_training_start(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Ran at the start of every training iteration. Required for optimizers."""
        raise NotImplementedError

    def on_training_end(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Ran at the end of every training iteration."""

    def on_testing_start(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Ran at the start of every testing iteration. This is initialized after """

    def on_testing_end(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Ran at the end of every testing iteration."""

    def on_validation_start(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Ran at the start of every validation iteration. By default, attempts to tune the learning rate."""
        parameters = model.parameters.copy()
        model._optimizer_lr *= 4
        self.on_training_start(model)
        loss_1 = model.loss(*model.training.sample)
        model.parameters = parameters
        model._optimizer_lr *= 0.125
        self.on_training_start(model)
        loss_2 = model.loss(*model.training.sample)
        model.parameters = parameters
        if loss_1 < loss_2:
            model._optimizer_lr *= 1.0625 / 0.5
        else:
            model._optimizer_lr *= 0.875 / 0.5
        model._optimizer_lr = max(min(model._optimizer_lr, self.lr * 8), self.lr / 8)


    def on_validation_end(self: Optimizer, model: EasyNN.model.abc.Model) -> None:
        """Ran at the end of every validation iteration."""

    def on_epoch_start(self: Optimizer, model: EasyNN.abc.Model) -> None:
        """Ran at the start of every epoch."""

    def on_epoch_end(self: Optimizer, model: EasyNN.abc.Model) -> None:
        """Ran at the end of every epoch."""



import hashlib
from PIL import Image
import io, os
import requests
import time
from selenium import webdriver

"""
Code downloaded from:
    Medium article: https://towardsdatascience.com/image-scraping-with-python-a96feda8af2d
    Author:Fabian Bosler
"""

def fetch_image_urls(query:str, max_links_to_fetch:int, wd:webdriver, sleep_between_interactions:int=1):
    def scroll_to_end(wd):
        wd.execute_script("window.scrollTo(0, document.body.scrollHeight);")
        time.sleep(sleep_between_interactions)    
    
    # build the google query
    search_url = "https://www.google.com/search?safe=off&site=&tbm=isch&source=hp&q={q}&oq={q}&gs_l=img"

    # load the page
    wd.get(search_url.format(q=query))

    image_urls = set()
    image_count = 0
    results_start = 0
    while image_count < max_links_to_fetch:
        scroll_to_end(wd)

        # get all image thumbnail results
        thumbnail_results = wd.find_elements_by_css_selector("img.Q4LuWd")
        number_results = len(thumbnail_results)
        
        print(f"Found: {number_results} search results. Extracting links from {results_start}:{number_results}")
        
        for img in thumbnail_results[results_start:number_results]:
            # try to click every thumbnail such that we can get the real image behind it
            try:
                img.click()
                time.sleep(sleep_between_interactions)
            except Exception as e:
                print(e)
                continue

            # extract image urls    
            actual_images = wd.find_elements_by_css_selector('img.n3VNCb')
            for actual_image in actual_images:
                if actual_image.get_attribute('src') and 'http' in actual_image.get_attribute('src'):
                    image_urls.add(actual_image.get_attribute('src'))

            image_count = len(image_urls)

            if len(image_urls) >= max_links_to_fetch:
                print(f"Found: {len(image_urls)} image links, done!")
                break
        else:
            print("Found:", len(image_urls), "image links, looking for more ...")
            time.sleep(30)
            #return
            load_more_button = wd.find_element_by_css_selector(".mye4qd")
            if load_more_button:
                wd.execute_script("document.querySelector('.mye4qd').click();")

        # move the result startpoint further down
        results_start = len(thumbnail_results)

    return image_urls

def persist_image(folder_path:str,url:str,index:int):
    try:
        image_content = requests.get(url).content

    except Exception as e:
        print(f"ERROR - Could not download {url} - {e}")

    try:
        image_file = io.BytesIO(image_content)
        image = Image.open(image_file).convert('RGB')
        file_path = os.path.join(folder_path, str(index) + '.jpg')
        with open(file_path, 'wb') as f:
            image.save(f, "JPEG", quality=85)
        print(f"SUCCESS - saved {url} - as {file_path}")
    except Exception as e:
        print(f"ERROR - Could not save {url} - {e}")

def scrape_google(search_term:str, driver_path='./chromedriver', target_path='./images', count=5):
    """
    Used for scraping images from google using the chromedriver.

    Requirement:
        chromedriver - 
            1. Install Google Chrome (skip if its already installed)
            2. Identify your Chrome version. In chrome go to  "About Google Chrome". I 
            currently have version 77.0.3865.90 (my main version is thus 77, the number
            before the first dot).
            
            https://chromedriver.chromium.org/downloads

            Download your corresponding ChromeDriver from the link provided for your main
            version and put the executable into the same folder as your python file.

    Function parameters:
        
        search_term: The term you want to scrape google images for.

        count: Number of image to scrape from google.

        driver_path: Defaulted to the path where the python file is executed.

        target_path: Creates a directory with target_path name and puts all
        the images inside here.

    Return:
        Images folder with sub folder of the search term with images inside of there.

    Example:
        >>> from EasyNN.scrape.scrape import scrape_google
        >>> scrape_google("dog")
        Opens browser and downloads 5 images of dogs and puts the into images folder.

        >>> scrape_google("dog", count=100)
        Opens browser and downloads 100 images of dog.

        >>> scrape_google("dog", target_path='./my_new_folder')
        Opens browser and downloads 5 images of dog and put it into a folder named "my_new_folder".
    """
    target_folder = os.path.join(target_path,'_'.join(search_term.lower().split(' ')))

    if not os.path.exists(target_folder):
        os.makedirs(target_folder)

    with webdriver.Chrome(executable_path=driver_path) as wd:
        res = fetch_image_urls(search_term, count, wd=wd, sleep_between_interactions=0.5)

    for index, elem in enumerate(res):
        persist_image(target_folder,elem,index)




"""
TODO: documentation.
"""
from __future__ import annotations
from abc import ABC
from typing import Any, Type


class AutoDocumentation(ABC):
    """
    Inherited by EasyNN.abc classes to automatically document implementation methods
    with the doc-string from the parent abstract method.
    """

    def __init_subclass__(cls: Type[AutoDocumentation], **kwargs: Any) -> None:
        """
        Initiallize subclasses by setting up the documentation for dunder methods
        which don't have any documentation by using the default documentation.
        """
        super().__init_subclass__(**kwargs)
        # Search for abstract methods.
        for super_cls in cls.mro():
            for name, attr in vars(super_cls).items():
                if getattr(attr, "__isabstractmethod__", False) and hasattr(cls, name) and getattr(cls, name).__doc__ is None:
                    getattr(cls, name).__doc__ = attr.__doc__



"""
TODO: Not complete.
"""
from __future__ import annotations
from EasyNN.optimizer.abc import Optimizer
from EasyNN.typing import ArrayND
import EasyNN.model.abc


class GradientDescent(Optimizer):
    """Gradient Descent uses parameters -= lr * derivatives."""

    def on_training_start(self: GradientDescent, model: EasyNN.model.abc.Model, derivatives: ArrayND = None) -> None:
        if derivatives is None:
            derivatives = self.get_derivatives(model)
        model.parameters -= model._optimizer_lr / (1 + 1e-2 * model.training.iteration) ** 0.3 * derivatives



import matplotlib.pyplot as plt
import numpy as np

file = "number_dataset.npz"
url =  "https://github.com/danielwilczak101/EasyNN/raw/datasets/mnist/number/number_dataset.npz"

labels = {
    0: 0,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
    6: 6,
    7: 7,
    8: 8,
    9: 9,
}

# Image related functions.
def show(user_image: list[list[int]], image_type: str = None) -> None:
    """Show mnist 28 by 28 images as either array data or a matplotlib image.
    
    Args:
        user_image: User image to be shown or the path of the image is handled in the exceptions
        image_type: Tells whether to show the image as an image or print out.
            image: Show as matplotlib image. (default use.)
            array: Show as numpy formated array print out.

    Returns:
        Either an matplotlib graphed image or numpy print of the image pixel values.

    Example:
        >>> show(model.training.data[0])
        Shows plotted image of data point
        >>> show(model.training.data[0],"array")
        Shows print out of properly spaved numpy array data.
    """

    if image_type is None or image_type == "image":
        try:
            plt.imshow(user_image.reshape((28, 28)), cmap='gray')
            plt.show()
        except ValueError:
            raise ValueError(f"Image is not sized to [28,28] or [1,784].\n \
            Try: image = Image(user_image).format(resized=[28,28])") from None
    elif image_type == "array":
        # Show the array with corrected print width.
        np.set_printoptions(linewidth=114)
        # Show the image array data.
        print(user_image)
        # Return the print to the default 75.
        np.set_printoptions(linewidth=75)


from EasyNN.model import Model
from EasyNN.utilities.download import download

# Required for the trained model.
download("fashion_parameters.npz","https://github.com/danielwilczak101/EasyNN/raw/datasets/mnist/fashion/fashion_parameters.npz")
download("fashion_structure.pkl","https://github.com/danielwilczak101/EasyNN/raw/datasets/mnist/fashion/fashion_structure.pkl")

# Load the saved model
model = Model.load("fashion")



from clint.textui import progress
from os.path import exists
import requests

__all__ = ["download"]

def download(file_name: str,  url: str) -> str:
    """Used for downloading dataset files to be used in the models.

    Args:
        url: Url to file to be downloaded.
        file_name: Name for file to be saved under.

    Returns:
        file_name: Returns the name of the file downloaded.
        File in the same directory where it is being run.

    Example:
        >>> from EasyNN.utilities.download import download
        >>> download("four.jpg","https://bit.ly/3lAJrMe")
        Downloading - four.jpg:
        [################################] 1768/1768 - 00:00:00
        
        File now stored in running file directory.
    """

    if not exists(file_name):
        # Tell the user what they are downloading.
        print(f"Downloading - {file_name}:")
        # Grab the file.
        r = requests.get(url, stream=True)
        # Produces the progress bar while downloading.
        with open(file_name, 'wb') as f:
            total_length = int(r.headers.get('content-length'))
            for chunk in progress.bar(r.iter_content(chunk_size=1024), expected_size=(total_length/1024) + 1):
                if chunk:
                    f.write(chunk)
                    f.flush()
    
    # Give the user the file name if they want to use it in anther function.
    return file_name







"""
TODO: Documentation.
"""
from __future__ import annotations
from itertools import accumulate
import numpy as np
from typing import Callable, Generic, Iterator, Optional, Union, TypeVar, overload
from EasyNN.model.abc import Model, base_array
from EasyNN.model.dense_layer import DenseLayer
from EasyNN.typing import Array1D, ArrayND

ArrayIn = TypeVar("ArrayIn", bound=ArrayND)
ArrayOut = TypeVar("ArrayOut", bound=ArrayND)


class Network(Model[ArrayIn, ArrayOut], Generic[ArrayIn, ArrayOut]):
    """
    A full Network model, which chains together several layers into one.

    If a layer is provided as an int, then base_layer(int) is used.
    Otherwise if a layer is provided as a Model, then that layer is used.
    Otherwise layer() is used e.g. ReLU -> ReLU().

    Example:
        >>> model = Network(128, ReLU, 10, SoftMax())
        >>> # 128 -> DenseLayer(128);  for an int input.
        >>> # ReLU -> ReLU();          for a class input.
        >>> # SoftMax() -> SoftMax();  for an already initialized layer.
    """
    __parameters: Array1D
    __derivatives: Array1D
    _layers: list[Model]

    def __init__(self: Network[ArrayIn, ArrayOut], *layers: Union[int, Model, Callable[[], Model]], base_layer: Callable[[int], Model] = DenseLayer) -> None:
        self._layers = [
            self.as_layer(layer, base_layer)
            for layer in layers
        ]

    def add(self: Network, layer: Union[int, Model, Callable[[], Model]], base_layer: Callable[[int], Model] = DenseLayer) -> None:
        """Append another layer to the end of the layers."""
        self._layers.append(self.as_layer(layer, base_layer))

    @staticmethod
    def as_layer(layer: Union[int, Model, Callable[[], Model]], base_layer: Callable[[int], Model] = DenseLayer) -> Model:
        """
        Parse the layer as an actual model.

        Examples
        --------
        >>> DenseNetwork.as_layer(128)        # DenseLayer(128);  for an int input.
        >>> DenseNetwork.as_layer(ReLU)       # ReLU();           for a class input.
        >>> DenseNetwork.as_layer(SoftMax())  # SoftMax();        for an already initialized layer.
        """
        # SoftMax() -> SoftMax().
        if isinstance(layer, Model):
            return layer
        # SoftMax -> SoftMax().
        elif callable(layer):
            return layer()
        # 128 -> DenseLayer(128)
        else:
            return base_layer(layer)

    @property
    def layers(self: Network[ArrayIn, ArrayOut]) -> tuple[Model, ...]:
        """Returns the layers of the network. Includes nested layers."""
        return tuple([
            sub_layer
            for layer in self._layers
            for sub_layer in layer.layers
        ])

    @property
    def _default_loss(self: Network[ArrayIn, ArrayOut]) -> Callable[[], Loss[ArrayIn, ArrayOut]]:
        """By default, use the loss that the last layer has by default (e.g. if LogLikelihood if the last layer is SoftMax)."""
        return self.layers[-1]._default_loss

    @property
    def _parameters(self: Network[ArrayIn, ArrayOut]) -> Array1D:
        """Provides views into the layers' parameters."""
        return self.__parameters

    @_parameters.setter
    def _parameters(self: Network[ArrayIn, ArrayOut], parameters: Array1D) -> None:
        if not all(hasattr(layer, "parameters") for layer in self.layers):
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute 'parameters' because a layer did not setup its parameters,\n"
                f"give the model some data to set itself up.\n"
                f"Example:\n"
                f">>> model(np.empty(28 * 28))     # Setting the input size."
                f">>> model(model.training[0][0])  # Using the training dataset."
            )
        self.__parameters = parameters
        i = 0
        for layer in self.layers:
            layer._parameters = parameters[i:i+len(layer.parameters)]
            i += len(layer.parameters)

    @property
    def _derivatives(self: Network[ArrayIn, ArrayOut]) -> Array1D:
        """Provides views into the layers' derivatives."""
        return self.__derivatives

    @_derivatives.setter
    def _derivatives(self: Network[ArrayIn, ArrayOut], derivatives: Array1D) -> None:
        self.__derivatives = derivatives
        i = 0
        for layer in self.layers:
            layer._derivatives = derivatives[i:i+len(layer.derivatives)]
            i += len(layer.derivatives)

    def get_arrays(self) -> dict[str, ArrayND]:
        """Returns the arrays stored in the model."""
        arrays = {
            f"{name}_{i}": arr
            for i, layer in enumerate(self.layers)
            for name, arr in layer.get_arrays().items()
            if name != "parameters"
        }
        if hasattr(self, "parameters"):
            arrays["parameters"] = self.parameters
        return arrays

    def set_arrays(self, *, parameters: ArrayND = None, **layer_arrays: ArrayND) -> None:
        """Sets the arrays stored in the model."""
        # Set the parameters, if any.
        if parameters is not None:
            self.parameters = parameters
        # Set the arrays for each individual layer.
        for i, layer in enumerate(self.layers):
            # Only use the arrays which match the current layer.
            layer.set_arrays(**{
                name.removesuffix(f"_{i}"): arr
                for name, arr in layer_arrays.items()
                if name.endswith(f"_{i}")
            })

    def __forward__(self: Network[ArrayIn, ArrayOut], x: ArrayIn) -> ArrayOut:
        # Pass the value through every layer.
        for layer in self._layers:
            x = layer(x)
        # After every layer is setup, setup the whole network.
        self.__setup__()
        return x

    def __backward__(self: Network[ArrayIn, ArrayOut], dy: ArrayOut, y: ArrayOut = None, use_y: bool = False):
        # Pass the derivative through every layer in reversed order.
        for layer in reversed(self._layers):
            dy = layer.backward(dy, y, use_y)
            # Only use y for the last layer, if at all.
            use_y = False
        return dy

    def __setup__(self: Network[ArrayIn, ArrayOut]) -> None:
        """Setup the Network parameter/derivative arrays as well as the arrays for each layer."""
        # Update the parameters every layer is setup,
        # but some are not setup to be the same as the DenseLayer parameters.
        if (
            all(hasattr(layer, "parameters") for layer in self.layers)
            and any(base_array(getattr(self, "parameters", None)) is not base_array(layer.parameters) for layer in self.layers)
        ):
            self._parameters = np.concatenate([layer.parameters for layer in self.layers], axis=0)
        # Update the derivatives every layer is setup,
        # but some are not setup to be the same as the DenseLayer derivatives.
        if (
            all(hasattr(layer, "derivatives") for layer in self.layers)
            and any(base_array(getattr(self, "derivatives", None)) is not base_array(layer.derivatives) is not None for layer in self.layers)
        ):
            self._derivatives = np.concatenate([layer.derivatives for layer in self.layers], axis=0)



from EasyNN.examples.mnist.fashion.structure import model
from EasyNN.examples.mnist.fashion.data import dataset
from EasyNN.examples.mnist.fashion import labels, show
from EasyNN.callbacks import ReachValidationAccuracy
from EasyNN.optimizer import MomentumDescent
import EasyNN.callbacks as cb
import numpy as np

# Assign it some training/testing data.
model.training.data = dataset

# Set when to terminate point. 
# In this case it will end once your validation accuracy hits above 90% three times.
model.callback(ReachValidationAccuracy(limit=0.90, patience=3))

# Used for plotting
model.validation_lr = 0.3
model.validation.accuracy = []

# Establish the labels and show feature.
model.labels = labels
model.show = show

# Use gradient descent with momentum.
model.optimizer = MomentumDescent()

model.callback(
    # Set when to terminate point. 
        # In this case it will end once your validation accuracy hits above 90% five times.
    cb.ReachValidationAccuracy(limit=0.30, patience=2),
)

# When the model hit a validation point it will print the iteration and accuracy of the model.
model.print.on_validation_start(iteration=True,accuracy=True)
# When the model completes 10 iterations. It will print that iteration number.
model.print.on_training_start(iteration=True, frequency=10)







import EasyNN.model.activation as activation
from EasyNN.model.abc import Model
from EasyNN.model.activation import *
from EasyNN.model.bias import Bias
from EasyNN.model.dense_layer import DenseLayer
from EasyNN.model.network import Network
from EasyNN.model.normalize import Normalize
from EasyNN.model.randomize import Randomize
from EasyNN.model.weight import Weight



from EasyNN.model import Model
from EasyNN.utilities.download import download

# Required for the trained model.
download("cifar10_parameters.npz", "https://github.com/danielwilczak101/EasyNN/raw/datasets/cifar/cifar10_parameters.npz")
download("cifar10_structure.pkl", "https://github.com/danielwilczak101/EasyNN/raw/datasets/cifar/cifar10_structure.pkl")

# Load the saved model
model = Model.load("cifar10")



"""
TODO: Documentation.
"""
from __future__ import annotations
import numpy as np
from typing import overload
from EasyNN.model.abc import Model_1D
from EasyNN.typing import Array1D


class Bias(Model_1D):
    neurons: int

    def __init__(self: Bias, neurons: int = None) -> None:
        if neurons is not None:
            self.neurons = neurons

    def __backward__(self: Bias, dy: Array1D, y: Array1D = None, use_y: bool = False) -> Array1D:
        if dy.shape != self.x.shape:
            raise ValueError(f"expected the same shape as the input, which is {self.x.shape}, not {dy.shape}")
        elif dy.ndim == 1:
            self.derivatives = dy
        elif dy.ndim == 2:
            self.derivatives = dy.mean(axis=0)
        else:
            raise ValueError(f"expected either 1-D or 2-D arrays, not {dy.ndim} arrays")
        return dy

    def __forward__(self: Bias, x: Array1D) -> Array1D:
        if x.ndim not in (1, 2):
            raise ValueError(f"expected either a 1-D or 2-D arrays, not a {dy.ndim}-D array")
        x += self.parameters
        return x

    def __setup__(self: Bias) -> None:
        # Get the number of neurons from the input.
        if not hasattr(self, "neurons"):
            self.neurons = self.x.shape[-1]
        # Setup the parameters and derivatives using the number of neurons.
        if not hasattr(self, "parameters"):
            self.parameters = np.zeros(self.neurons, dtype=float)
        if not hasattr(self, "derivatives"):
            self.derivatives = np.empty(self.neurons, dtype=float)



from __future__ import annotations
from copy import copy
from EasyNN.typing import ArrayND
from typing import Generic, TypeVar

Array = TypeVar("Array", bound=ArrayND[float])


class Momentum(Generic[Array]):
    """Basic implementation of an exponentially moving average used for momentum."""
    _weight: float = 0.0
    _value: Array = 0.0
    lr: float

    @property
    def value(self: Momentum[Array]) -> Array:
        return self._value / self._weight

    @value.setter
    def value(self: Momentum[Array], value: Array) -> None:
        self._weight += self.lr * (1 - self._weight)
        self._value += self.lr * (value - self._value)

    def __init__(self: Momentum[Array], lr: float = 1e-2) -> None:
        self.lr = lr

    def update(self: Momentum[Array], value: Array) -> Array:
        """Update and return the momentum value."""
        self.value = value
        return self.value

    def peek(self: Momentum[Array], value: Array) -> Array:
        """Peek at the result of `momentum.update(value)` without updating the momentum."""
        _weight = copy(self._weight)
        _value = copy(self._value)
        value = self.update(value)
        self._weight = _weight
        self._value = _value
        return value



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='EasyNN',
    version='0.1.3',
    description='EasyNN is a python package designed to provide an easy-to-use neural network. The package is designed to work right out of the box, while also allowing the user to customize features as they see fit.',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    url="https://github.com/danielwilczak101/EasyNN",
    author="Daniel Wilczak, Jack Nguyen",
    author_email="danielwilczak101@gmail.com",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    classifier=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: OS Independent",
        ],
    install_requires = ["matplotlib ~= 3.3.2",
                        "clint >=0.5.1",
                        "tabulate >=0.8.7",
                        "requests >=2.26.0",
                        "numpy >=1.21.3",
                        "nptyping >=1.4.4",
                        "selenium>=4.1.0"
                        ],
    )



import numpy as np

def load(file_path: str):
    """Used for loading dataset files that have been downloaded.
    
    Args:
        file_path: Path to file to be loaded.
        
    Returns:
        x: Data used to train models.
        y: Dataset labels.

   Example:
        >>> data,labels = load("model/mnist.npz")
        >>> # Print first dataset example and first label
        >>> print(data[0])
        >>> print(label[0])
        [0 200 ... 15  0]
        5
    """
    
    with np.load(file_path) as data:
        return  data['x'], \
                data['y']



"""
TODO: Not complete.
"""
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any
from EasyNN._abc import AutoDocumentation
from EasyNN.classifier.classifier import Classifier
from EasyNN.typing import ArrayND, Factory
import EasyNN.model.abc


class Loss(AutoDocumentation, ABC):
    """
    Abstract Base Class for the Loss, which measures the performance of a model,
    provided batches of data. The Loss is also used to compute derivatives,
    which are backpropagated into the model.

    This class is used between the Model and the Optimizer.

    Example
    -------
    >>> from EasyNN.batch.mini import MiniBatch
    >>> model = ...
    >>> model.loss = ...
    >>> batch = MiniBatch(...)
    >>> x, y_true = batch.sample
    >>> y_pred = model(x)
    >>> accuracy = model.loss(*batch.sample)
    >>> model.backward(model.loss.dy(*batch.sample))
    """
    model: EasyNN.model.abc.Model
    _default_classifier: Factory[Classifier] = Classifier

    @abstractmethod
    def __call__(self: Loss, x: Array1D, y: Array1D) -> float:
        """
        Calling loss(x, y) will provide the the loss value.

        Parameters
        ----------
        sample: Sample = self.training.sample
            Uses the sample to compute the loss on.

        Returns
        -------
        loss_value: float
            A measurement of how good the model is. This value should be minimized.

        Example
        -------
        >>> training_accuracy = loss(*training.sample)
        >>> testing_accuracy = loss(*testing.sample)
        """
        raise NotImplementedError

    def backward(self: Loss, y: ArrayND, y_pred: ArrayND, model: EasyNN.model.abc.Model[ArrayND, ArrayND] = None) -> Any:
        """
        Updates self.derivatives (e.g. via backpropagation) for optimizers (e.g. gradient descent).
        Default implementation returns model.backward(loss.dy(y, model(x))).

        Parameters:
            y: Array1D
                The expected output of the model.
            y_pred: Array1D
                The predicted output from the model.
            model: Model
                The model being backpropagated into.

        Example:
            >>> model.backward_sample(x, y)           # For the user.
            >>> model.backward(loss.dy(y, model(x)))  # How it's implemented.
        """
        if model is None:
            model = self.model
        return model.backward(self.dy(y, y_pred))

    @abstractmethod
    def dy(self: Loss, y: ArrayND, y_pred: ArrayND) -> ArrayND:
        """
        Computes the change in y expected, which is used to backpropagate into the model.
        Usually involves loss.y_pred and loss.y_true.

        Parameters:
            y, y_pred: Array1D
                The expected and predicted y values.

        Returns:
            dy: ArrayND
                The change in y that should be backpropagated into the model.
        """
        raise NotImplementedError



from EasyNN.model import Network, Normalize, Randomize, ReLU, LogSoftMax

# Create the Neural Network Model.
model = Network(
    Normalize(1e-3),
    Randomize(0.5), 800, ReLU,
    Randomize(0.1), 10,  LogSoftMax
)